@@ -4,32 +4,30 @@ import json
 import re
 import sys
 import uuid
-from datetime import datetime
-from datetime import timedelta
+from datetime import datetime, timedelta
 from unittest import mock
-from urllib.parse import quote
-from urllib.parse import urlparse
-from urllib.parse import urlunparse
+from urllib.parse import quote, urlparse, urlunparse
 
-from pytest import fixture
-from pytest import mark
+from pytest import fixture, mark
 from tornado.httputil import url_concat
 
 import jupyterhub
+
 from .. import orm
 from ..apihandlers.base import PAGINATION_MEDIA_TYPE
 from ..objects import Server
 from ..utils import url_path_join as ujoin
 from ..utils import utcnow
 from .conftest import new_username
-from .mocking import public_host
-from .mocking import public_url
-from .utils import add_user
-from .utils import api_request
-from .utils import async_requests
-from .utils import auth_header
-from .utils import find_user
-
+from .utils import (
+    add_user,
+    api_request,
+    async_requests,
+    auth_header,
+    find_user,
+    public_host,
+    public_url,
+)
 
 # --------------------
 # Authentication tests
@@ -471,6 +469,42 @@ async def test_get_users_state_filter(app, state):
     assert usernames == expected
 
 
+@mark.user
+async def test_get_users_name_filter(app):
+    db = app.db
+
+    add_user(db, app=app, name='q')
+    add_user(db, app=app, name='qr')
+    add_user(db, app=app, name='qrs')
+    add_user(db, app=app, name='qrst')
+    added_usernames = {'q', 'qr', 'qrs', 'qrst'}
+
+    r = await api_request(app, 'users')
+    assert r.status_code == 200
+    response_users = [u.get("name") for u in r.json()]
+    assert added_usernames.intersection(response_users) == added_usernames
+
+    r = await api_request(app, 'users?name_filter=q')
+    assert r.status_code == 200
+    response_users = [u.get("name") for u in r.json()]
+    assert response_users == ['q', 'qr', 'qrs', 'qrst']
+
+    r = await api_request(app, 'users?name_filter=qr')
+    assert r.status_code == 200
+    response_users = [u.get("name") for u in r.json()]
+    assert response_users == ['qr', 'qrs', 'qrst']
+
+    r = await api_request(app, 'users?name_filter=qrs')
+    assert r.status_code == 200
+    response_users = [u.get("name") for u in r.json()]
+    assert response_users == ['qrs', 'qrst']
+
+    r = await api_request(app, 'users?name_filter=qrst')
+    assert r.status_code == 200
+    response_users = [u.get("name") for u in r.json()]
+    assert response_users == ['qrst']
+
+
 @mark.user
 async def test_get_self(app):
     db = app.db
@@ -1389,6 +1423,17 @@ async def test_get_proxy(app):
     assert list(reply.keys()) == [app.hub.routespec]
 
 
+@mark.parametrize("offset", (0, 1))
+async def test_get_proxy_pagination(app, offset):
+    r = await api_request(
+        app, f'proxy?offset={offset}', headers={"Accept": PAGINATION_MEDIA_TYPE}
+    )
+    r.raise_for_status()
+    reply = r.json()
+    assert set(reply) == {"items", "_pagination"}
+    assert list(reply["items"].keys()) == [app.hub.routespec][offset:]
+
+
 async def test_cookie(app):
     db = app.db
     name = 'patience'
@@ -2068,14 +2113,23 @@ def test_shutdown(app):
         )
         return r
 
-    real_stop = loop.stop
+    real_stop = loop.asyncio_loop.stop
 
     def stop():
         stop.called = True
         loop.call_later(1, real_stop)
 
-    with mock.patch.object(loop, 'stop', stop):
+    real_cleanup = app.cleanup
+
+    def cleanup():
+        cleanup.called = True
+        return real_cleanup()
+
+    app.cleanup = cleanup
+
+    with mock.patch.object(loop.asyncio_loop, 'stop', stop):
         r = loop.run_sync(shutdown, timeout=5)
     r.raise_for_status()
     reply = r.json()
+    assert cleanup.called
     assert stop.called
