@@ -8,10 +8,9 @@ import pytest
 from pytest import raises
 from traitlets.config import Config
 
-from ..app import JupyterHub
-from ..app import NewToken
-from ..app import UpgradeDB
-
+from .. import orm
+from ..app import NewToken, UpgradeDB
+from ..scopes import _check_scopes_exist
 
 here = os.path.abspath(os.path.dirname(__file__))
 populate_db = os.path.join(here, 'populate_db.py')
@@ -36,7 +35,9 @@ def generate_old_db(env_dir, hub_version, db_url):
     check_call([env_py, populate_db, db_url])
 
 
-@pytest.mark.parametrize('hub_version', ['0.7.2', '0.8.1', '0.9.4'])
+# changes to this version list must also be reflected
+# in ci/init-db.sh
+@pytest.mark.parametrize('hub_version', ["1.1.0", "1.2.2", "1.3.0", "1.5.0", "2.1.1"])
 async def test_upgrade(tmpdir, hub_version):
     db_url = os.getenv('JUPYTERHUB_TEST_DB_URL')
     if db_url:
@@ -75,3 +76,10 @@ async def test_upgrade(tmpdir, hub_version):
 
     # run tokenapp again, it should work
     tokenapp.start()
+
+    db = orm.new_session_factory(db_url)()
+    query = db.query(orm.APIToken)
+    assert query.count() >= 1
+    for token in query:
+        assert token.scopes, f"Upgraded token {token} has no scopes"
+        _check_scopes_exist(token.scopes)
